use anyhow::Result;
use camino::{Utf8Path, Utf8PathBuf};
use chrono::{DateTime, Duration, Utc};
use feed_rs::model::Entry;
use feed_rs::model::Feed;
use ron::ser::{to_string_pretty, PrettyConfig};
use serde::{Deserialize, Serialize};
use std::collections::{BTreeMap, HashMap};
use std::convert::AsRef;
use std::fs;
use std::time::Instant;
use ureq::http::HeaderMap;
use ureq::http::Response;
use ureq::Body;
use url::Url;

pub fn slugify_url(url: &Url) -> String {
    let domain = url.domain().unwrap();
    let query = url.query().unwrap_or("");
    slug::slugify(format!("{domain}{}{query}", url.path()))
}

/// Stored settings/info about a feed.
#[derive(Debug, Deserialize, Serialize)]
pub struct FeedStoreFeedInfo {
    /// First time we added this feed.
    ///
    /// Used for historical purposes only.
    pub added: DateTime<Utc>,
    /// Last known cached entry, if any.
    ///
    /// Used to let the server know whether we need a new entry or not.
    pub fetch_data: Option<FetchData>,
}

impl FeedStoreFeedInfo {
    pub fn new() -> Self {
        Self {
            added: Utc::now(),
            fetch_data: None,
        }
    }
}

/// Storage for a single feed.
///
/// Contains one [FeedStoreVersion] for every time the feed has been successfully fetched,
/// and one [FeedStoreEntry] for each article referenced throughout the entries.
#[derive(Debug, Deserialize, Serialize)]
pub struct FeedStoreFeed {
    /// The feed URL
    pub url: Url,
    /// Where it's stored, should be inside the [FeedStore::dir].
    pub dir: Utf8PathBuf,
    /// Raw feed path
    pub path_feed: Utf8PathBuf,
    /// Raw feed RON path
    pub path_feed_ron: Utf8PathBuf,
    /// Settings path
    pub path_settings: Utf8PathBuf,
    /// Detailed settings/info about a feed.
    pub info: FeedStoreFeedInfo,
    /// Stored copy of the raw XML feed (if any)
    pub raw_feed: Option<String>,
}

impl FeedStoreFeed {
    pub fn new(basedir: &Utf8Path, url: &Url) -> Self {
        let dir = basedir.join(slugify_url(url));
        if !dir.is_dir() {
            std::fs::create_dir_all(&dir).unwrap();
        }

        let path_settings = dir.join("settings.toml");
        let info: FeedStoreFeedInfo = match std::fs::read_to_string(&path_settings) {
            Ok(s) => toml::from_str(&s).unwrap(),
            Err(_e) => {
                // Assume file has not been created yet. Initialize
                let info = FeedStoreFeedInfo::new();
                std::fs::write(&path_settings, toml::to_string(&info).unwrap()).unwrap();
                info
            }
        };

        let raw_feed: Option<String> = std::fs::read_to_string(dir.join("feed.xml")).ok();

        Self {
            dir: dir.clone(),
            path_feed: dir.join("feed.xml"),
            path_feed_ron: dir.join("feed.ron"),
            path_settings: dir.join("settings.toml"),
            url: url.clone(),
            info,
            raw_feed,
        }
    }

    pub fn load_fetchdata(&self) -> Option<&FetchData> {
        self.info.fetch_data.as_ref()
    }

    pub fn load_feed(&self, sanitize: bool) -> Option<Feed> {
        if let Some(raw_feed) = &self.raw_feed {
            let parser = feed_rs::parser::Builder::new()
                .sanitize_content(sanitize)
                .build();
            Some(parser.parse(raw_feed.as_bytes()).unwrap())
        } else {
            None
        }
    }

    pub fn has_changed(&self, new_feed: &Feed) -> bool {
        let Some(old_feed) = self.load_feed(false) else {
            return true;
        };

        let mut old_iter = old_feed.entries.iter();
        for new in &new_feed.entries {
            let Some(old) = old_iter.next() else {
                return true;
            };
            if old != new {
                return true;
            }
        }
        // ignoring any entries left in old_iter
        false
    }

    pub fn store(&mut self, mut response: Response<Body>) -> Result<bool> {
        let headers = response.headers();
        let fetchdata = FetchData {
            etag: hv(headers, "etag"),
            last_modified: hv(headers, "last_modified"),
            when: Utc::now(),
        };

        let body = response.body_mut().with_config().read_to_vec()?;
        let feed = match feed_rs::parser::parse(body.as_slice()) {
            Ok(f) => f,
            Err(e) => {
                warn!("Error when parsing feed for {}: {e:?}", self.url);
                return Ok(false);
            }
        };

        debug!("Storing fetchdata for {}", self.url);
        self.info.fetch_data = Some(fetchdata);
        Self::write(&self.path_settings, toml::to_string(&self.info)?)?;

        if !self.has_changed(&feed) {
            return Ok(false);
        }
        debug!("Storing feed for {}.", self.url);
        // todo don't serialize to string but to writer
        Self::write(
            &self.path_feed_ron,
            to_string_pretty(&feed, PrettyConfig::default())?,
        )?;
        Self::write(&self.path_feed, body)?;
        Ok(true)
    }

    /// refresh in hours
    pub fn fetch(&mut self, fetcher: &super::Fetcher, refresh: usize) -> Result<bool> {
        let mut builder = fetcher
            .agent
            .get(self.url.to_string())
            .header("FROM", fetcher.from.clone());

        if let Some(fetchdata) = self.load_fetchdata() {
            if !fetchdata.etag.is_empty() {
                builder = builder.header("If-None-Match", fetchdata.etag.clone());
            }
            if !fetchdata.last_modified.is_empty() {
                builder = builder.header("If-Modified-Since", fetchdata.last_modified.clone());
            }

            // Check if we have hit time for refresh
            if fetchdata.when + Duration::try_hours(refresh as i64).unwrap() >= Utc::now() {
                // No need to rebuild, check again later
                return Ok(false);
            }
        }

        let start_instant = Instant::now();
        let result = builder.call();
        let duration = start_instant.elapsed();

        let response = result?;
        debug!(
            "fetched with status {} in {} ms: {}",
            response.status(),
            duration.as_millis(),
            self.url,
        );
        let status = response.status();
        match status.as_u16() {
            304 => Ok(false), // Not Modified -> nothing to do
            200 => self.store(response),
            _ => {
                warn!(
                    "HTTP Status {} not implemented for {}",
                    response.status(),
                    self.url,
                );
                Ok(false)
            }
        }
    }

    fn write<P: AsRef<std::path::Path> + std::fmt::Display, C: AsRef<[u8]>>(
        path: P,
        contents: C,
    ) -> std::io::Result<()> {
        if fs::exists(&path)? {
            fs::rename(&path, format!("{path}.backup"))?;
        }
        fs::write(path, contents)
    }
}

#[derive(Clone, Debug, Deserialize, Serialize, Default)]
pub struct FetchData {
    pub when: DateTime<Utc>,
    pub etag: String,
    pub last_modified: String,
}

#[derive(Debug)]
pub struct FeedStore {
    pub _dir: Utf8PathBuf,
    pub feeds: BTreeMap<Url, FeedStoreFeed>,
}

impl FeedStore {
    pub fn new(dir: &str, feedlist: &Vec<super::FeedConfig>) -> Self {
        let dir = super::to_checked_pathbuf(dir);
        let mut feeds: BTreeMap<Url, FeedStoreFeed> = BTreeMap::new();

        for feed_config in feedlist {
            let feed_url = Url::parse(&feed_config.url).unwrap();
            feeds.insert(feed_url.clone(), FeedStoreFeed::new(&dir, &feed_url));
        }

        Self { _dir: dir, feeds }
    }

    pub fn collect(&mut self, max_entries: usize) -> (HashMap<String, Feed>, Vec<Entry>) {
        let mut feeds = HashMap::new();
        let mut entries = Vec::new();

        for (feed_url, feed_store_feed) in self.feeds.iter_mut() {
            let Some(mut feed) = feed_store_feed.load_feed(true) else {
                warn!("Problem parsing feed file for feed {}", feed_url);
                continue;
            };

            for entry in &mut feed.entries {
                entry.source = Some(feed_url.to_string());
            }

            entries.append(&mut std::mem::take(&mut feed.entries));
            if entries.len() > 4 * max_entries {
                entries = trim_entries(entries, max_entries);
            }

            feeds.insert(feed_url.to_string(), feed.clone());
        }
        (feeds, trim_entries(entries, max_entries))
    }

    pub fn fetch(&mut self, fetcher: &super::Fetcher, refresh: usize) -> Result<bool> {
        let mut rebuild = false;
        for (_url, feed) in self.feeds.iter_mut() {
            rebuild |= feed.fetch(fetcher, refresh)?;
        }

        Ok(rebuild)
    }
}

fn trim_entries(mut entries: Vec<Entry>, max_entries: usize) -> Vec<Entry> {
    entries.sort_by_key(|e| std::cmp::Reverse(e.updated.or(e.published).unwrap_or_default()));
    entries.truncate(max_entries);
    entries
}

fn hv(headers: &HeaderMap, key: &str) -> String {
    match headers.get(key) {
        Some(hv) => hv.to_str().unwrap_or_default().to_string(),
        _ => "".to_string(),
    }
}
